from confluent_kafka import Producer

from resistant_kafka.producer_schemas import ProducerConfig, DataSend


class ProducerInitializer:
    def __init__(
            self,
            config: ProducerConfig,
    ):
        self._producer_name = config.producer_name
        self._producer = Producer(self._set_producer_config(config=config))

    @staticmethod
    def _set_producer_config(config: ProducerConfig) -> dict:
        producer_config = {
            'bootstrap.servers': config.bootstrap_servers,
        }
        if config.secured:
            producer_config['oauth_cb'] = config.oauth_cb
            producer_config['security.protocol'] = config.security_protocol
            producer_config['sasl.mechanisms'] = config.sasl_mechanisms

        return producer_config

    @staticmethod
    def _delivery_report(error_message, message):
        if error_message is not None:
            print("Delivery failed for User record {}: {}".format(message.key(), error_message))
            return
        print(
            "User record {} successfully produced to {} [{}] at offset {}".format(
                message.key(),
                message.topic(),
                message.partition(),
                message.offset()
            )
        )

    def send_message(
            self,
            data_to_send: DataSend,
    ):
        self._producer.produce(
            topic=self._producer_name,
            key=data_to_send.key,
            value=data_to_send.value,
            on_delivery=self._delivery_report,
            headers=data_to_send.headers
        )
        self._producer.flush()



from dataclasses import dataclass
from typing import Callable, Any, Union

from pydantic import BaseModel, model_validator


class ProducerConfig(BaseModel):
    """
        bootstrap_servers: str -- This attribute uses for connection to Kafka server, where user try to connect
        producer_name: str -- This attribute uses for connection to Kafka server, where user try to connect

    """
    bootstrap_servers: str
    producer_name: str
    secured: bool
    oauth_cb: Union[Callable, None] = None
    security_protocol: Union[str, None] = None
    sasl_mechanisms: Union[str, None] = None

    @model_validator(mode='after')
    def check_secured_config(self):
        if self.secured:
            missing_fields = []
            if self.oauth_cb is None:
                missing_fields.append("oauth_cb")

            if self.security_protocol is None:
                missing_fields.append("security_protocol")

            if self.sasl_mechanisms is None:
                missing_fields.append("sasl_mechanisms")

            if missing_fields:
                raise ValueError(
                    f"If 'secured' is True, the following fields must be set: {', '.join(missing_fields)}"
                )
        return self


@dataclass
class DataSend:
    key: str
    value: Any
    headers: list[tuple] | None = None



import asyncio
import functools
import logging
from abc import abstractmethod
from typing import Any

from confluent_kafka import Consumer

from resistant_kafka.common_exceptions import KafkaMessageError
from resistant_kafka.consumer_schemas import ConsumerConfig

logging.basicConfig(level=logging.INFO)


class ConsumerInitializer:
    def __init__(
            self,
            config: ConsumerConfig
    ):
        self._consumer = Consumer(
            self._set_consumer_config(config=config)
        )
        self._consumer.subscribe(
            topics=[config.topic_to_subscribe],
            on_assign=self._connection_flag_method
        )
        self._config = config

    @staticmethod
    def _set_consumer_config(config: ConsumerConfig) -> dict:
        consumer_config = {
            'bootstrap.servers': config.bootstrap_servers,
            'group.id': config.group_id,
            'auto.offset.reset': config.auto_offset_reset,
            'enable.auto.commit': config.enable_auto_commit
        }
        if config.secured:
            consumer_config['oauth_cb'] = config.oauth_cb
            consumer_config['security.protocol'] = config.security_protocol
            consumer_config['sasl.mechanisms'] = config.sasl_mechanisms

        return consumer_config

    def _connection_flag_method(self, *args):
        logging.info(f"{self._config.processor_name} successful subscribed "
                     f"to the topic {self._config.topic_to_subscribe}\n")

    @staticmethod
    def message_is_empty(message: Any, consumer: Consumer):
        if message is None:
            consumer.commit(asynchronous=True)
            return True

        if getattr(message, "key", None) is None:
            consumer.commit(asynchronous=True)
            return True

        if message.key() is None:
            consumer.commit(asynchronous=True)
            return True

        return False

    @staticmethod
    async def get_message(consumer):
        loop = asyncio.get_running_loop()
        poll = functools.partial(consumer.poll, 1.0)
        return await loop.run_in_executor(executor=None, func=poll)

    @abstractmethod
    async def process(self):
        pass


def kafka_processor(raise_error=False):
    def handle_kafka_errors(func):
        async def wrapper(self, *args, **kwargs):
            while True:
                try:
                    await func(self, *args, **kwargs)
                except Exception as e:
                    if raise_error:
                        raise KafkaMessageError(str(e))

                    print(f"Kafka processing error: {e}")

                finally:
                    self._consumer.commit(asynchronous=True)

        return wrapper

    return handle_kafka_errors


async def process_kafka_connection(tasks: list[ConsumerInitializer]):
    while True:
        await asyncio.gather(*[task.process() for task in tasks])


def init_kafka_connection(tasks: list[ConsumerInitializer]):
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)

    loop.create_task(process_kafka_connection(tasks=tasks))

    loop.run_forever()



from .consumer import ConsumerInitializer, init_kafka_connection, kafka_processor
from .consumer_schemas import ConsumerConfig
from .producer import ProducerInitializer
from .producer_schemas import ProducerConfig, DataSend



from typing import Union, Callable

from pydantic import BaseModel, model_validator


class ConsumerConfig(BaseModel):
    """
        topic_to_subscribe: str -- Kafka topic, that specific processor will process
        processor_name: str -- This name can be random, just for logging
        bootstrap_servers: str -- This attribute uses for connection to Kafka server, where user try to connect
        group_id: str -- Consumer name, which would be shown as subscribed to specific topic
        auto_offset_reset: Literal ['latest', 'earliest'] -- When the group is first created, before any
                messages have been consumed, the position is set according to a configurable offset reset policy
        enable_auto_commit: bool -- If True: Kafka Consumer automatically commits message offsets
                at a specified time interval (every 5 seconds).
                                    If False: Ensure that offset is only saved after the message
                has been successfully processed.
        secured: bool -- Uses to determine if it required to get JWT token for security
        oauth_cb: Callable | None -- Function object, that uses for getting token
        security_protocol: str
        sasl_mechanisms: str
    """
    topic_to_subscribe: str
    processor_name: str
    bootstrap_servers: str
    group_id: str
    auto_offset_reset: str = 'latest'
    enable_auto_commit: bool = True
    secured: bool = False
    oauth_cb: Union[Callable, None] = None
    security_protocol: Union[str, None] = None
    sasl_mechanisms: Union[str, None] = None

    @model_validator(mode='after')
    def check_auto_offset_reset_config(self):
        if self.auto_offset_reset not in ['latest', 'earliest']:
            raise ValueError(
                f"Auto offset reset attribute have to have only latest or earliest"
            )

    def check_secured_config(self):
        if self.secured:
            missing_fields = []
            if self.oauth_cb is None:
                missing_fields.append("oauth_cb")

            if self.security_protocol is None:
                missing_fields.append("security_protocol")

            if self.sasl_mechanisms is None:
                missing_fields.append("sasl_mechanisms")

            if missing_fields:
                raise ValueError(
                    f"If 'secured' is True, the following fields must be set: {', '.join(missing_fields)}"
                )
        return self



from setuptools import setup, find_packages

setup(
    name='resistant_kafka',
    version='0.1',
    packages=find_packages(),
    install_requires=[
        "confluent-kafka==2.8.2",
        "dotenv==0.9.9",
        "requests==2.32.3",
        "pydantic==2.11.3"
    ]
)
